/*
 * libdigidocpp
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include "Conf.h"

#include "log.h"
#include "SignatureBES.h"
#include "crypto/Digest.h"
#include "crypto/X509Cert.h"
#include "util/File.h"

#include <map>

using namespace digidoc;
using namespace digidoc::util;
using namespace std;

Conf* Conf::INSTANCE = nullptr;

Conf::Conf()
{
}

Conf::~Conf()
{
}

Conf* Conf::instance()
{
	return INSTANCE;
}

/**
 * Init global Conf with conf
 * @param conf implementation to use
 **/
void Conf::init(Conf *conf)
{
    delete INSTANCE;
    INSTANCE = conf;
}

bool Conf::bdoc1Supported() const
{
    return false;
}

string Conf::defaultPolicyId() const
{
    return SignatureBES::policylist.cbegin()->first;
}

string Conf::libdigidocConf() const
{
#ifdef _WIN32
    return File::dllPath("digidoc.dll") + "digidoc.ini";
#else
    return "";
#endif
}

int Conf::logLevel() const {
#ifdef NDEBUG
    return Log::InfoType;
#else
    return Log::DebugType;
#endif
}
string Conf::logFile() const { return ""; }
string Conf::digestUri() const { return URI_SHA256; }

string Conf::xsdPath() const
{
    string path = "schema";
#ifdef __APPLE__
    path.insert(0, File::frameworkResourcesPath("ee.ria.digidocpp"));
#endif
    return path;
}

string Conf::PKCS11Driver() const { return PKCS11_MODULE; }

string Conf::ocsp(const string &issuer) const
{
    static const map<string,string> ocsplist = [](){
        map<string,string> list;
        //Estonia Live
        list.insert(make_pair("ESTEID-SK 2007", "http://ocsp.sk.ee"));
        list.insert(make_pair("ESTEID-SK 2011", "http://ocsp.sk.ee"));
        list.insert(make_pair("KLASS3-SK 2010", "http://ocsp.sk.ee"));
        //Estonia Test
        list.insert(make_pair("TEST of ESTEID-SK 2007", "http://www.openxades.org/cgi-bin/ocsp.cgi"));
        list.insert(make_pair("TEST of ESTEID-SK 2011", "http://www.openxades.org/cgi-bin/ocsp.cgi"));
        list.insert(make_pair("TEST of KLASS3-SK 2010", "http://www.openxades.org/cgi-bin/ocsp.cgi"));
        //Finland Test
        list.insert(make_pair("VRK CA for Test Purposes", "http://www.openxades.org/cgi-bin/ocsp.cgi"));
        list.insert(make_pair("VRK CA for Test Purposes - G2", "http://www.openxades.org/cgi-bin/ocsp.cgi"));
        //Latvia Test - disabled, issuer name is identical with live certificates
        //list.insert(make_pair("E-ME SI (CA1)", "http://www.openxades.org/cgi-bin/ocsp.cgi"));
        //Lithuania Test
        list.insert(make_pair("Nacionalinis sertifikavimo centras (IssuingCA A)", "http://www.openxades.org/cgi-bin/ocsp.cgi"));
        return list;
    }();
    auto pos = ocsplist.find(issuer);
    return pos == ocsplist.end() ? "http://ocsp.sk.ee/_proxy" : pos->second;
}

string Conf::certsPath() const
{
#ifdef __APPLE__
    string path = File::frameworkResourcesPath("ee.ria.digidocpp");
    if(!path.empty())
        return path + "certs";
#endif
#ifdef CERTS_PATH
    return CERTS_PATH;
#else
    return "";
#endif
}

string Conf::proxyHost() const { return ""; }
string Conf::proxyPort() const { return ""; }
string Conf::proxyUser() const { return ""; }
string Conf::proxyPass() const { return ""; }
string Conf::PKCS12Cert() const
{
#ifdef __APPLE__
    string path = File::frameworkResourcesPath("ee.ria.digidocpp");
#elif defined(_WIN32) && defined(_DEBUG)
    string path = File::dllPath("digidocppd.dll");
#elif defined(_WIN32)
    string path = File::dllPath("digidocpp.dll");
#else
    string path = DIGIDOCPP_CONFIG_DIR "/";
#endif
    return path + "37242.p12";
}
string Conf::PKCS12Pass() const { return "\x30\x75\x35\x61\x52\x55\x6b\x41"; }
bool Conf::PKCS12Disable() const { return false; }

ConfV2::ConfV2() {}
ConfV2::~ConfV2() {}
ConfV2* ConfV2::instance() { return dynamic_cast<ConfV2*>(Conf::instance()); }

string ConfV2::TSUrl() const { return TSA_URL; }



bool ConfV2::TSLAutoUpdate() const { return true; }
string ConfV2::TSLCache() const
{
#ifdef _WIN32
    string cachePath = File::env("APPDATA");
    return cachePath += "\\digidocpp\\tsl\\";
#else
    string cachePath = File::env("HOME");
    return cachePath += "/.digidocpp/tsl/";
#endif
}
X509Cert ConfV2::TSLCert() const
{
    const unsigned char cert[] = {
        "-----BEGIN CERTIFICATE-----\n"
        "MIIHPDCCBSSgAwIBAgIBcjANBgkqhkiG9w0BAQsFADAxMQswCQYDVQQGEwJFUzERMA8GA1\n"
        "UECgwIRk5NVC1SQ00xDzANBgNVBAMMBklTQSBDQTAeFw0xMTExMTYxNTExMDdaFw0xNTEx\n"
        "MTYxNTExMDdaME0xCzAJBgNVBAYTAkJFMRwwGgYDVQQKDBNFVVJPUEVBTiBDT01NSVNTSU\n"
        "9OMSAwHgYDVQQDDBcoU0lHTikgQU5ORUxJIEFORFJFU1NPTjCCASIwDQYJKoZIhvcNAQEB\n"
        "BQADggEPADCCAQoCggEBAMw881fbxBn5Eq64Hf6pqEqn0Intg49k5D+E2JBMcyeOMYE4Ww\n"
        "uzVqs7fuKMX5mhiFw7llVx42S7GIDnKa6ascF7irISsYTGOtD6dGt8OEOQkG5i24wxT3we\n"
        "63VW1W/PEO0WoRjkPDBqgljnqcLn+WJLr3Yl9fpkxyT+H3pl0jnGkGi7UUKEDuoCglFy9T\n"
        "b1TE+eVNiGWmfskRy6q9Pioujp2FqrwNm95VAkLFGoNF4iZQda79bgjvAYnA5p4QXgr/5s\n"
        "WZjgsa3Ea9PlRwEvO7C8ndnRE/PTUNwaSmOZsmPH/ksCCnfTDqqehHVnry3K6QIUKhRsoT\n"
        "1m60qv4UKrvrsCAwEAAaOCA0EwggM9MF0GA1UdEQRWMFSBHUFOTkVMSS5BTkRSRVNTT05A\n"
        "RUMuRVVST1BBLkVVpDMwMTEYMBYGCSsGAQQBrGYBAgwJQU5EUkVTU09OMRUwEwYJKwYBBA\n"
        "GsZgEBDAZBTk5FTEkwCQYDVR0TBAIwADAOBgNVHQ8BAf8EBAMCBkAwHQYDVR0OBBYEFEoU\n"
        "Ef+0Oe59fVRs/rW2M0KhmlDsMB8GA1UdIwQYMBaAFEft+GPwma9e/n4OXFjL/uI1N6a9MI\n"
        "HgBgNVHSAEgdgwgdUwgcgGCisGAQQBrGYDBAEwgbkwKQYIKwYBBQUHAgEWHWh0dHA6Ly93\n"
        "d3cuY2VydC5mbm10LmVzL2RwY3MvMIGLBggrBgEFBQcCAjB/DH1RdWFsaWZpZWQgY2VydG\n"
        "lmaWNhdGUuIFVuZGVyIHRoZSB1c2FnZSBjb25kaXRpb25zIGFzc2VydGVkIGluIHRoZSBG\n"
        "Tk1ULVJDTSBDUFMgKDEwNiwgSm9yZ2UgSnVhbiBzdHJlZXQsMjgwMDksIE1hZHJpZCwgU3\n"
        "BhaW4pLjAIBgYEAIswAQEwgYYGCCsGAQUFBwEBBHoweDBBBggrBgEFBQcwAYY1aHR0cDov\n"
        "L29jc3BJU0FjYS5jZXJ0LmZubXQuZXMvb2NzcElTQWNhL09jc3BSZXNwb25kZXIwMwYIKw\n"
        "YBBQUHMAKGJ2h0dHA6Ly93d3cuY2VydC5mbm10LmVzL2NlcnRzL0lTQUNBLmNydDBGBggr\n"
        "BgEFBQcBAwQ6MDgwCAYGBACORgEBMAsGBgQAjkYBAwIBDzAVBgYEAI5GAQIwCxMDRVVSAg\n"
        "ECAgECMAgGBgQAjkYBBDCBzAYDVR0fBIHEMIHBMIG+oIG7oIG4hoGIbGRhcDovL2xkYXBJ\n"
        "U0FjYS5jZXJ0LmZubXQuZXMvQ049Q1JMMixjbj1JU0ElMjBDQSxvPUZOTVQtUkNNLEM9RV\n"
        "M/Y2VydGlmaWNhdGVSZXZvY2F0aW9uTGlzdDtiaW5hcnk/YmFzZT9vYmplY3RjbGFzcz1j\n"
        "UkxEaXN0cmlidXRpb25Qb2ludIYraHR0cDovL3d3dy5jZXJ0LmZubXQuZXMvY3Jsc19JU0\n"
        "FjYS9DUkwyLmNybDANBgkqhkiG9w0BAQsFAAOCAgEALIKumRPBo9NIYyzm4LSJklE9Cmsi\n"
        "rJC89MPcFiC9+vg4YrO9Mmakr8G5ZAVZf+Zr6HgLcIf93/BSgE4UDFD9mAROPDElqx2ppg\n"
        "1+foIvlqg57EiHH2kOS4IYJEJpWCs4IqcBkSKcS5vJ5XVpNJL4elpeMQjsGoqjmkKDr1Ie\n"
        "ZQVOxOiW/bNhM9Fw6R1aMMLbnziwAgHWDpmIj2Se74HuS3ca0ctm9fAPp/2Y87bgZ1ETs9\n"
        "A6wc4GxYEuGTwh1HPbDB6tgjSnjH0hY3qBITjB8IM0xGvUR9dQkt8Ez4dvyxcaFlyCMfLa\n"
        "eMf8jlJOv4uU1yRRT9wESIj5GcoE43cy1cuvRMifqYhoi3KPojJaLTV7ASy1X0yuBWyg0j\n"
        "GyvE6aTIAu3XGCwdGYB1WnSzEUPJboqcyJydkXXxnBKFchw602pVB0ITUqitGH+Kfix2mQ\n"
        "Qf+v0RDEWQU3WdADNHc+cqoc7sXkJXkO657dYZT/Ey5MR+gXRm6oK52iXIqjM8kbDqCqAO\n"
        "HVC73u+bUvh0y5B5PjGYuE2K9xK36ooavh5ytF/kY+y0Ju804CRF00b4wKcibGxplOrU1g\n"
        "+ncONNmBHGbsLULXrum8+fsD3O04Mf0t0O9Mt/ZwAZTa8Cwq/0pqOWdRqucV7pq/A7fEnR\n"
        "b5AdBaAzxqrXVaa92NkScAm80=\n"
        "-----END CERTIFICATE-----\n"
    };
    return X509Cert(vector<unsigned char>(cert, cert + sizeof(cert)), X509Cert::Pem);
}
string ConfV2::TSLUrl() const
{ return "https://ec.europa.eu/information_society/policy/esignature/trusted-list/tl-mp.xml"; }
