#include "Connect.h"

#include "Conf.h"
#include "Container.h"
#include "crypto/OpenSSLHelpers.h"

#include <openssl/bio.h>
#include <openssl/ocsp.h>
#include <openssl/ssl.h>

using namespace digidoc;
using namespace std;

Connect::Connect(const string &url, const string &method)
    : senteol(false)
{
    char *_host = 0, *_port = 0, *_path = 0;
    int usessl = 0;
    if(!OCSP_parse_url(const_cast<char*>(url.c_str()), &_host, &_port, &_path, &usessl))
        THROW_OPENSSLEXCEPTION("Incorrect URL provided: '%s'.", url.c_str());

    chost = host = _host ? _host : "";
    cport = port = _port ? _port : "80";
    path = _path ? _path : "/";
    OPENSSL_free(_host);
    OPENSSL_free(_port);
    OPENSSL_free(_path);

    Conf *c = Conf::instance();
    if(!c->proxyHost().empty())
    {
        chost = c->proxyHost();
        cport = c->proxyPort();
    }

    string hostname = chost + ":" + cport;
    SCOPE2(BIO, bio, BIO_new_connect(const_cast<char*>(hostname.c_str())), BIO_free_all);
    if(!bio)
        THROW_OPENSSLEXCEPTION("Failed to create connection with host: '%s'", hostname.c_str());

    if(usessl > 0)
    {
        ssl.reset(SSL_CTX_new(SSLv23_client_method()), function<void(SSL_CTX*)>(SSL_CTX_free));
        if(!ssl)
            THROW_OPENSSLEXCEPTION("Failed to create connection with host: '%s'", hostname.c_str());
        SSL_CTX_set_mode(ssl.get(), SSL_MODE_AUTO_RETRY);
        BIO *sbio = BIO_new_ssl(ssl.get(), 1);
        if(!sbio)
            THROW_OPENSSLEXCEPTION("Failed to create ssl connection with host: '%s'", hostname.c_str());
        d.reset(BIO_push(sbio, bio.release()), function<void(BIO*)>(BIO_free_all));
    }
    else
        d.reset(bio.release(), function<void(BIO*)>(BIO_free_all));

    if(!d)
        THROW_OPENSSLEXCEPTION("Failed to create connection with host: '%s'", hostname.c_str());

    if(!BIO_do_connect(d.get()))
        THROW_OPENSSLEXCEPTION("Failed to connect to host: '%s'", hostname.c_str());

    BIO_printf(d.get(), "%s %s HTTP/1.0\r\n", method.c_str(), path.c_str());
    BIO_printf(d.get(), "Host: %s:%s\r\n", host.c_str(), port.c_str());
    BIO_printf(d.get(), "User-Agent: LIB libdigidocpp/%s APP %s\r\n",
        VER_STR(MAJOR_VER.MINOR_VER.RELEASE_VER.BUILD_VER), appInfo().c_str());

    if(!c->proxyUser().empty() || !c->proxyPass().empty())
    {
        BIO *b64 = BIO_new(BIO_f_base64());
        BIO_set_flags(b64, BIO_FLAGS_BASE64_NO_NL);
        SCOPE2(BIO, hash, BIO_push(b64, BIO_new(BIO_s_mem())), BIO_free_all);
        BIO_printf(hash.get(), "%s:%s", c->proxyUser().c_str(), c->proxyPass().c_str());
        (void)BIO_flush(hash.get());
        char *base64 = 0;
        BIO_get_mem_data(hash.get(), &base64);
        BIO_printf(d.get(), "Proxy-Authorization: Basic %s\r\n", base64);
    }
}

void Connect::addContent(const vector<unsigned char> &data)
{
    BIO_printf(d.get(), "Content-Length: %ld\r\n\r\n", data.size());
    BIO_write(d.get(), &data[0], int(data.size()));
    senteol = true;
}

void Connect::addHeader(const string &key, const string &value)
{
    BIO_printf(d.get(), "%s: %s\r\n", key.c_str(), value.c_str());
}

string Connect::read() const
{
    if(!senteol)
        BIO_printf(d.get(), "\r\n");

    int rc = 0;
    size_t pos = 0;
    string result(1024, 0);
    do {
        if(size_t(pos += rc) >= result.size())
            result.resize(result.size()*2);
    } while((rc = BIO_read(d.get(), &result[pos], int(result.size() - pos))) > 0);
    result.resize(pos);
    pos = result.find("\r\n\r\n");
    return pos == string::npos ? result : result.substr(pos + 4);
}
